/*
 * Copyright 2008, Systemic Realtime Design LLC. All Rights Reserved.
 *
 * This source code may contain confidential information of Systemic
 * Realtime Design LLC and its licensors. Any use, reproduction,
 * modification, disclosure, distribution or transfer of this software,
 * or any software which includes or is based upon any of this code, is
 * prohibited unless expressly authorized by SRD by written agreement. For
 * more information (including whether this source code file has been
 * published) please email licensing@sysrtime.com.
 */


#ifndef TIPC_SOCKET_H
#define TIPC_SOCKET_H

/* includes */
#include <inttypes.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/socketvar.h>
//#include "sys/queue.h"
#include "tipc_skbuff.h"


#define TIPC_TX_BUFSIZE (256<<10U)
#define TIPC_RX_BUFSIZE (1024<<10U)
extern int optb;

/* TIPC protocol control block (PCB) */
		
struct sock {			
	int state;                    /* BSD struct socket uses so_state as a bitflag so TIPC state is here */
	clock_t sk_rcvtimeo;          /* time limit for ACK (in ms) */
	int (*sk_backlog_rcv)(struct sock *sk, struct sk_buff *skb);
	void *sk_user_data;           /* user data: tipc port for socket */
	pthread_mutex_t sk_lock;      /* owner lock */
	uint32_t sk_sleep;            /* wakes up would-be owners */
	pthread_t owner;              /* thread id of lock owner */
	uint32_t waiters;             /* # tasks waiting for socket ownership */
	int need_pollwakeup;
	struct socket *sk_socket;     /* pointer to generic socket */
	struct {
	    struct sk_buff *head;     /* first message in deferred queue */
	    struct sk_buff *tail;     /* last message in deferred queue */
	} sk_backlog;
};


#endif  /* TIPC_SOCK_H */

